@@ -26,6 +26,10 @@ var http = require('http')
   , CookieJar = require('./vendor/cookie/jar')
   , cookieJar = new CookieJar
   ;
+  
+if (process.logging) {
+  var log = process.logging('request')
+}
 
 try {
   https = require('https')
@@ -151,7 +155,13 @@ Request.prototype.request = function () {
     // fetch cookie from the global cookie jar
     var cookies = cookieJar.get({ url: self.uri.href })
   }
-  if (cookies) {self.headers.Cookie = cookies}
+  if (cookies) {
+    var cookieString = cookies.map(function (c) {
+      return c.name + "=" + c.value;
+    }).join("; ");
+    
+    self.headers.Cookie = cookieString;
+  }
 
   if (!self.uri.pathname) {self.uri.pathname = '/'}
   if (!self.uri.port) {
@@ -198,8 +208,8 @@ Request.prototype.request = function () {
     } 
     if (!form) form = {}
     var oa = {}
-    for (i in form) oa[i] = form[i]
-    for (i in self.oauth) oa['oauth_'+i] = self.oauth[i]
+    for (var i in form) oa[i] = form[i]
+    for (var i in self.oauth) oa['oauth_'+i] = self.oauth[i]
     if (!oa.oauth_version) oa.oauth_version = '1.0'
     if (!oa.oauth_timestamp) oa.oauth_timestamp = Math.floor( (new Date()).getTime() / 1000 ).toString()
     if (!oa.oauth_nonce) oa.oauth_nonce = uuid().replace(/-/g, '')
@@ -215,7 +225,7 @@ Request.prototype.request = function () {
     var signature = oauth.hmacsign(self.method, baseurl, oa, consumer_secret, token_secret)
     
     // oa.oauth_signature = signature
-    for (i in form) {
+    for (var i in form) {
       if ( i.slice(0, 'oauth_') in self.oauth) {
         // skip 
       } else {
@@ -223,8 +233,8 @@ Request.prototype.request = function () {
       }
     }
     self.headers.authorization = 
-      'OAuth '+Object.keys(oa).sort().map(function (i) {return i+'="'+encodeURIComponent(oa[i])+'"'}).join(',')
-    self.headers.authorization += ',oauth_signature="'+encodeURIComponent(signature)+'"'  
+      'OAuth '+Object.keys(oa).sort().map(function (i) {return i+'="'+oauth.rfc3986(oa[i])+'"'}).join(',')
+    self.headers.authorization += ',oauth_signature="'+oauth.rfc3986(signature)+'"'  
   }
 
   if (self.uri.auth && !self.headers.authorization) {
@@ -253,7 +263,7 @@ Request.prototype.request = function () {
     }
 
   } else if (self.multipart) {
-    self.body = ''
+    self.body = []
     self.headers['content-type'] = 'multipart/related;boundary="frontier"'
     if (!self.multipart.forEach) throw new Error('Argument error, options.multipart.')
 
@@ -261,28 +271,44 @@ Request.prototype.request = function () {
       var body = part.body
       if(!body) throw Error('Body attribute missing in multipart.')
       delete part.body
-      self.body += '--frontier\r\n'
+      var preamble = '--frontier\r\n'
       Object.keys(part).forEach(function(key){
-        self.body += key + ': ' + part[key] + '\r\n'
+        preamble += key + ': ' + part[key] + '\r\n'
       })
-      self.body += '\r\n' + body + '\r\n'
+      preamble += '\r\n'
+      self.body.push(new Buffer(preamble))
+      self.body.push(new Buffer(body))
+      self.body.push(new Buffer('\r\n'))
     })
-    self.body += '--frontier--'
+    self.body.push(new Buffer('--frontier--'))
   }
 
   if (self.body) {
+    var length = 0
     if (!Buffer.isBuffer(self.body)) {
-      self.body = new Buffer(self.body)
+      if (Array.isArray(self.body)) {
+        for (var i = 0; i < self.body.length; i++) {
+          length += self.body[i].length
+        }
+      } else {
+        self.body = new Buffer(self.body)
+        length = self.body.length
+      }
+    } else {
+      length = self.body.length
     }
-    if (self.body.length) {
-      self.headers['content-length'] = self.body.length
+    if (length) {
+      self.headers['content-length'] = length
     } else {
       throw new Error('Argument error, options.body.')
     }
   }
 
-  self.httpModule =
-    {"http:":http, "https:":https}[self.proxy ? self.proxy.protocol : self.uri.protocol]
+  var protocol = self.proxy ? self.proxy.protocol : self.uri.protocol
+    , defaultModules = {'http:':http, 'https:':https}
+    , httpModules = self.httpModules || {}
+    ;
+  self.httpModule = httpModules[protocol] || defaultModules[protocol]
 
   if (!self.httpModule) throw new Error("Invalid protocol")
 
@@ -304,7 +330,8 @@ Request.prototype.request = function () {
   self.start = function () {
     self._started = true
     self.method = self.method || 'GET'
-
+    self.href = self.uri.href
+    if (log) log('%method %href', self)
     self.req = self.httpModule.request(self, function (response) {
       self.response = response
       response.request = self
@@ -319,6 +346,13 @@ Request.prototype.request = function () {
 
       if (setHost) delete self.headers.host
       if (self.timeout && self.timeoutTimer) clearTimeout(self.timeoutTimer)
+      
+      if (response.headers['set-cookie'] && (!self._disableCookies)) {
+        response.headers['set-cookie'].forEach(function(cookie) {
+          if (self.jar) self.jar.add(new Cookie(cookie))
+          else cookieJar.add(new Cookie(cookie))
+        })
+      }
 
       if (response.statusCode >= 300 &&
           response.statusCode < 400  &&
@@ -336,14 +370,18 @@ Request.prototype.request = function () {
           response.headers.location = url.resolve(self.uri.href, response.headers.location)
         }
         self.uri = response.headers.location
-        self.redirects.push( { statusCode : response.statusCode,
-                               redirectUri: response.headers.location })
+        self.redirects.push(
+          { statusCode : response.statusCode
+          , redirectUri: response.headers.location 
+          }
+        )
         delete self.req
         delete self.agent
         delete self._started
         if (self.headers) {
           delete self.headers.host
         }
+        if (log) log('Redirect to %uri', self)
         request(self, self.callback)
         return // Ignore the rest of the response
       } else {
@@ -412,7 +450,11 @@ Request.prototype.request = function () {
                 chunk.copy(body, i, 0, chunk.length)
                 i += chunk.length
               })
-              response.body = body.toString()
+              if (self.encoding === null) {
+                response.body = body
+              } else {
+                response.body = body.toString()
+              }
             } else if (buffer.length) {
               response.body = buffer.join('')
             }
@@ -422,17 +464,6 @@ Request.prototype.request = function () {
                 response.body = JSON.parse(response.body)
               } catch (e) {}
             }
-            if (response.statusCode == 200 && response.headers['set-cookie'] && (!self._disableCookies)) {
-              response.headers['set-cookie'].forEach(function(cookie) {
-                if (self.jar) {
-                  // custom defined jar
-                  self.jar.add(new Cookie(cookie));
-                } else {
-                  // add to the global cookie jar if user don't define his own
-                  cookieJar.add(new Cookie(cookie));
-                }
-              });
-            }
 
             self.callback(null, response, response.body)
           })
@@ -442,13 +473,13 @@ Request.prototype.request = function () {
 
     if (self.timeout) {
       self.timeoutTimer = setTimeout(function() {
-          self.req.abort()
-          var e = new Error("ETIMEDOUT")
-          e.code = "ETIMEDOUT"
-          self.emit("error", e)
+        self.req.abort()
+        var e = new Error("ETIMEDOUT")
+        e.code = "ETIMEDOUT"
+        self.emit("error", e)
       }, self.timeout)
     }
-
+    
     self.req.on('error', clientErrorHandler)
   }
 
@@ -478,7 +509,13 @@ Request.prototype.request = function () {
 
   process.nextTick(function () {
     if (self.body) {
-      self.write(self.body)
+      if (Array.isArray(self.body)) {
+        self.body.forEach(function(part) {
+          self.write(part)
+        })
+      } else {
+        self.write(self.body)
+      }
       self.end()
     } else if (self.requestBodyStream) {
       console.warn("options.requestBodyStream is deprecated, please pass the request object to stream.pipe.")
